TOOL_NAME = "Song Preview Sample"

--extra lua files
------------------
--require "general_functions"


--Renoise Keybinds and menus
-------------------------------
renoise.tool():add_keybinding {
  name = "Global:Tools:"..TOOL_NAME,
  invoke = function()main_toggle()end  
}

renoise.tool():add_menu_entry {
  name = "Main Menu:Tools:Ledger`s Scripts:"..TOOL_NAME,
  invoke = function()main_start()end  
}
--[[
renoise.tool():add_menu_entry {
  name = "Instrument Box:"..TOOL_NAME,
  invoke = function()check_for_preview_sample()end  
}
--]]
renoise.tool():add_menu_entry {
  name = "Instrument Box:".."Set Selected As #Song Preview Sample",
  invoke = function()set_current_instrument_as_preview_sample()end  
}


renoise.tool():add_keybinding {
  name = "Global:Tools:".."Set Selected As #Song Preview Sample",
  invoke = function()set_current_instrument_as_preview_sample()end  
}


--? TODO
--Always check/ move to top slot
--limit archived samples
--auto normalise/ level dB


------------------------------------------
--Set up Preferences file
------------------------------------------
--create xml
local options = renoise.Document.create {
  autoseek = false,
  us_time = false,
  time_stamp = false,
  reminder_interval = 5, 
}
 --assign options-object to .preferences so renoise knows to load and update it with the tool
renoise.tool().preferences = options
------------------------------------------
------------------------------------------
--variable syntax 
--options.value_a.value

--[[
--------------------------------------------------------------------------------
--some basic colors for gui elements
--------------------------------------------------------------------------------
--e.g. For changing vb.views["sample present colour 2"].color when states change
COLOR_GREY = {0x30,0x42,0x42}
COLOR_GREY_MILD = {0x60,0x75,0x75}
COLOR_ORANGE ={0xFF,0x66,0x00}
COLOR_YELLOW = {0xE0,0xE0,0x00}
COLOR_BLUE = {0x50,0x40,0xE0}  
COLOR_RED = {0xEE,0x10,0x10}
COLOR_GREEN = {0x20,0x99,0x20}
COLOR_RED_MILD = {0x90,0x10,0x10}

--Constants holding first index of color-code: Used to identify color in vb.views table i.e.
--if vb.views["button"].color[1] == COLOR_ORANGE_FLAG then --etc
COLOR_GREY_FLAG = COLOR_GREY[1]
COLOR_GREY_MILD = COLOR_GREY_MILD[1]
COLOR_ORANGE_FLAG = COLOR_ORANGE[1]
COLOR_YELLOW_FLAG = COLOR_YELLOW[1]
COLOR_BLUE_FLAG = COLOR_BLUE[1]  
COLOR_RED_FLAG = COLOR_RED[1]
COLOR_GREEN_FLAG = COLOR_GREEN[1]
COLOR_RED_MILD_FLAG = COLOR_RED_MILD[1]
--]]

LUA_COUNTS_FROM_1 = 1
INSTRUMENT_BASE_NAME = "#Song Preview"


--global variables for gui
local my_dialog = nil 
local vb = nil


--toggle the tool open and closed (keyboard shortcut start-up)
-------------------------------------------------------------
function main_toggle()
----------------------
 --close dialog if it is open
  if (my_dialog and my_dialog.visible) then 
    my_dialog:close()
    --reset global my_dialog
     my_dialog = nil 
  else --run main
    main()
  end
end

--always open/ restart tool (menu entry start-up)
-------------------------------------------------
function main_start()
---------------------
  if (my_dialog and my_dialog.visible) then 
    my_dialog:close()
    --reset global my_dialog
     my_dialog = nil 
  end
  --run main
  main()
end
----

--helper function : custom status message 
--prefixes tool name and adds brackets 
---------------------------------------
local function status(message)
------------------------------
  renoise.app():show_status(TOOL_NAME.." Tool: ("..message..")")
end

--\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

--returns the Preview Instrument or nil 
---------------------------------------
local function get_preview_instrument()
---------------------------------------
 local song = renoise.song()
 for i = 1,#song.instruments do
    if string.find(song:instrument(i).name,INSTRUMENT_BASE_NAME) ~= nil then 
      return song:instrument(i)
    end
  end
  return nil
end

--"Set Selected As #Song Preview Sample"
-----------------------------------------------------
function set_current_instrument_as_preview_sample()
-----------------------------------------------------
  
  local song = renoise.song()
  local new_inst = song.selected_instrument
  local new_inst_idx = song.selected_instrument_index
  local active_inst = nil
  local active_inst_idx = nil
  
  ----------------------------------------
  local function get_the_date(us_format)
  ----------------------------------------
    --get the date
    local time = os.time()
    local date_text = os.date("%x", time)--.."  "..os.date("%X", time)
  
    if us_format == false then
      local month =  string.sub(date_text,1,2) --10/02
      local day = string.sub(date_text,4,5)
      local year = string.sub(date_text,7,8)
     
      return day.."/"..month.."/"..year
    end
    --U.S. format
    return date_text
  end
  ----
  -------------------------------
  local function get_time_stamp()
  -------------------------------
    local time = os.time()
    local hour = os.date("%I", time)
    local mins = os.date("%M", time)
    local am_pm = os.date("%p", time)  
    
    return hour..":"..mins.." "..am_pm
  end
  
  local time_stamp = ""
  if options.time_stamp.value == true then
    time_stamp =" "..get_time_stamp()
  end
  
  ------------------------------
  local function get_file_name()
  ------------------------------
    local song = renoise.song()
    --Get the saved songs file name from the full save path 
    local file_name = song.file_name
    
    if file_name == "" then
      return "song not saved yet"
    end
    
    local length = string.len(file_name)
    --reverse so we can search for last backslash (Windows)
    file_name = string.reverse(file_name)
    --find "\"
    local last_back_slash = string.find(file_name,'\\')--length))--- #file_name))
    --reverse again back to original
    file_name = string.reverse(file_name)
    --last backslash pos (adjusted by 2) 
    last_back_slash = length - last_back_slash  + 2
    --get the wanted  substring out of the original path 
    return  string.sub(file_name,last_back_slash)
  end
  ---
  
  --TODO Normalise?
  
  for i = 1,#song.instruments do
    if string.find(song:instrument(i).name,INSTRUMENT_BASE_NAME) ~= nil then 
      active_inst = song:instrument(i)
      active_inst_idx = i
      break
    end
  end
  
  if active_inst_idx == song.selected_instrument_index then
    status("Need to Choose A New Render In Instrument Box To Add To "..INSTRUMENT_BASE_NAME)
    return
  end
    
  --Preview Sample instrument found so
  if active_inst ~= nil then
    --make sure the new instrument that we will be copying from has sample data in slot 1
    if (#new_inst.samples == 0) or (new_inst:sample(1).sample_buffer.has_sample_data == false) then
      status("No Valid Sample Present To Add To "..INSTRUMENT_BASE_NAME)
      return
    end
    --inserting a new sample at top to hold the new render (also archives previous renders further down the list)
    active_inst:insert_sample_at(1)
    --copy the sample (users render)
    active_inst:sample(1):copy_from(new_inst:sample(1))
    --rename the sample
    active_inst:sample(1).name = get_the_date(options.us_time.value)..time_stamp.." : "..get_file_name()    
    --rename instrument name that got lost in the copy operation above
    active_inst.name = INSTRUMENT_BASE_NAME.." [0]"

    --map samples to the white keys in downward direction (newest at C-4, oldest as lowest key)
    --------------------------------------------------------------------------------------------
    ---add to get white keys up the octaves
    --local a_keys = {0,2,4,5,7,9,11}
    --add to get white keys down the octaves
    local m_keys = {0,-1,-3,-5,-7,-8,-10}
    local note = 48 --c-4 = 48
    local m_keys_counter = 1
    for i = 1,#active_inst.samples do
      --adjust note to white keys
      local adjusted_note = note + m_keys[m_keys_counter]
      active_inst.samples[i].sample_mapping.note_range = {adjusted_note,adjusted_note} 
      active_inst.samples[i].sample_mapping.base_note = adjusted_note
      --check reached next octave
      if m_keys_counter == 7 then
        m_keys_counter = 0
        note = note - 12
        --no more keys to map so break mapping loop
        if note == 0 then
          break
        end
      end
      m_keys_counter = m_keys_counter + 1
    end
    
    --autoseek
    -----------
    for i = 1,#active_inst.samples do
      if options.autoseek.value == true then
        active_inst.samples[i].autoseek = true
      else
        active_inst.samples[i].autoseek = false
      end
    end
    
    --delete the render that we just copied from 
    song:delete_instrument_at(new_inst_idx)
    --focus active inst 
    song.selected_instrument_index = active_inst_idx
    --show sample editor
    renoise.app().window.active_middle_frame = renoise.ApplicationWindow.MIDDLE_FRAME_INSTRUMENT_SAMPLE_EDITOR  
  else
    --else set the currently selected instrument as the #Song Preview Sample
    if #new_inst.samples > 0 then
       new_inst.name = INSTRUMENT_BASE_NAME.." [0]"
      --get the date, pass in user preference for euro or U.S. date format
      local date_text = get_the_date(true)
      --update sample name
      local new_sample_name = date_text.." : "..get_file_name()
      new_inst:sample(1).name = new_sample_name
      status(INSTRUMENT_BASE_NAME.."Updated With Renamed Sample: "..new_sample_name)
    else
      status("No Valid Sample Present To Create "..INSTRUMENT_BASE_NAME)
    end
  end
end

---notifier fn
-----------------------------------
function check_for_preview_sample()
-----------------------------------

  --user has set to no reminders
  if options.reminder_interval.value == 0 then
    return
  end
  
  local song = renoise.song()
  local found_idx = false
  local reminder_num_reached = false
  
  
  
  local REMINDER_INTERVAL = options.reminder_interval.value
  --search for instrument  
  for i = 1,#song.instruments do
    if string.find(song:instrument(i).name,INSTRUMENT_BASE_NAME) ~= nil then
      found_idx = i
      
      --check if bypassed
      if song:instrument(i).name == INSTRUMENT_BASE_NAME.." [B]" or
        song:instrument(i).name == INSTRUMENT_BASE_NAME.." [b]" then
        status("BYPASSED")
        return
      end
      
      --increment counter
      local current_count = string.match(song:instrument(i).name,"%d+")
      --convert to number
      current_count = tonumber(current_count)

      if (current_count % REMINDER_INTERVAL) == 0 and (current_count ~= 0) then --minus 1 on current count, as we count from 1 with name counter
        reminder_num_reached = true
        break
      end
      --covert back to string
      current_count = tostring(current_count + 1) --increment
      song:instrument(i).name = INSTRUMENT_BASE_NAME.." ["..current_count.."]"
      break
    end
  end
  
  if reminder_num_reached == true then
     local text = INSTRUMENT_BASE_NAME.."' Tool:\n\nReminder To Update Your Preview Sample!\n\n(You Can Change the Frequency Of These Reminders In The Tool's Preferences)\n"
     local response = renoise.app():show_prompt(TOOL_NAME,text, {"OK", "Skip", "Prefs."})
     if response == "Show Prefs." then
       --main shows preferences window
       main()
       return
     elseif response == "Skip" then 
       song:instrument(found_idx).name = INSTRUMENT_BASE_NAME.." [0]"
      return
    end
  end
      
  if found_idx == false then
    local reply = renoise.app():show_prompt(INSTRUMENT_BASE_NAME, "'"..INSTRUMENT_BASE_NAME.."' Tool:\n\nNo Preview Render Slot Found", {"Add","Bypass Song","Ignore"})
    if reply == "Add" then
      local new_inst = song:insert_instrument_at(1)
      new_inst.name = INSTRUMENT_BASE_NAME.." [0]"
      return
    end

    if reply == "Bypass Song" then --add instrument with [B] for bypass
      local new_inst = song:insert_instrument_at(1)
      new_inst.name = INSTRUMENT_BASE_NAME.." [b]"
      return
    end
  end
end


--notifier on save to check for and then increment counter
if renoise.tool().app_saved_document_observable:has_notifier(check_for_preview_sample) == false then
 renoise.tool().app_saved_document_observable:add_notifier(check_for_preview_sample)
end

--renoise.tool().app_saved_document_observable



------------------------------------
------------------------------------
function main()
------------------------------------
------------------------------------


  --toggle dialog so if its already open then close it.  The timer will catch
  --that it is closed and do the housekeeping like removing notifiers (and timer itself)
  if (my_dialog and my_dialog.visible) then 
    my_dialog:close()
    return
  end
  
  --renoise song object
  local song = renoise.song()

  --Always check/ move to top slot
  --reminder (x)saves
  --date format U.S or Euro
  --limit archived samples
  --auto normalise/ level dB




  --set globals
  -------------
  --assign global vb to viewbuilder
  vb = renoise.ViewBuilder()
  
 
  ----------------------------------------------
  --GUI
  ----------------------------------------------
  local dialog_content = vb:column{
  width = 250,
   vb:row{ --row
   margin = 4,
   vb:text{
     text = "Autoseek Samples On:   ",
    },
   vb:checkbox{
     value = options.us_time.value,
     notifier = function(value)
                  --set option
                  options.autoseek.value = value
                  --get preview instrument
                  local preview_inst = get_preview_instrument()
                  if preview_inst == nil then
                    return
                  end
                  --set autoseek values in preview instrument samples
                  for i = 1,#preview_inst.samples do
                    if options.autoseek.value == true then
                      preview_inst.samples[i].autoseek = true
                    else
                      preview_inst.samples[i].autoseek = false
                    end
                  end
                end
    
    }
   },
  
  
  
  vb:row{ --row
   margin = 4,
   vb:text{
     text = "U.S. Date Notation:   ",
    },
   vb:checkbox{
     value = options.us_time.value,
     notifier = function(value)
                  options.us_time.value = value
                end
    
    }
   },
  vb:row{
   margin = 4,
   vb:text{
     text = "Show Timestamp:   ",
    },
   vb:checkbox{
     value = options.time_stamp.value,
     notifier = function(value)
                  options.time_stamp.value = value
                end
    
    }
   },
   
   vb:row{ 
    margin = 4,
    vb:text{
     text = "Song Saves Until Reminder (0 = OFF):",
    }
    },
   vb:row{ 
    margin = 4,
    vb:valuebox{
     min = 0,
     max = 20,
     value = options.reminder_interval.value,
      notifier = function(value)
                  options.reminder_interval.value = value
                end
    
   } 
  }
 }     




  --------------
  --key Handler
  --------------
  local function my_keyhandler_func(dialog,key)
     --toggle lock focus hack, allows pattern ed to get key input
     renoise.app().window.lock_keyboard_focus = not renoise.app().window.lock_keyboard_focus
     renoise.app().window.lock_keyboard_focus = not renoise.app().window.lock_keyboard_focus
    -- rprint(key)
     
     return key
     
  end 
  
 
  ------------------------------------------------------------------------------------------
  ------------------------------------------------------------------------------------------
  --Script dialog
  my_dialog = renoise.app():show_custom_dialog(
      TOOL_NAME.." Options", dialog_content,my_keyhandler_func)

 --[[ 
  --add timer to fire once every 50ms
  if not renoise.tool():has_timer(timer) then
    renoise.tool():add_timer(timer,80)
  end
  --]]
  
  --------------------------------------------    
  --close dialog function ON NEW SONG
  --------------------------------------------
  local function closer(d)
    --close dialog if exists and is open     
    if (d ~= nil) and (d.visible == true) then
      d:close()
     -- remove_notifiers()
    end
    --reset global my_dialog
     my_dialog = nil 
  end
  -- notifier to close dialog on load new song
  renoise.tool().app_release_document_observable:add_notifier(closer,my_dialog)
  -------------------------------------------------------------------------------
  
  --first run of timer
  -- timer()
end

--[[
----------------
--timer function
----------------
function timer()
  
  --do stuff
  
  --remove timer when GUI is closed
  if(my_dialog == nil) or (my_dialog.visible == false) then
    if renoise.tool():has_timer(timer) then
      renoise.tool():remove_timer(timer)
    end
  end
end
--]]



